local tomQuest2 = LibStub("AceAddon-3.0"):NewAddon("tomQuest2", "AceEvent-3.0", "AceBucket-3.0")
-----------------------
--- Library loading ---
-----------------------
local LC = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-Core")
local BC = LibStub("LibBabble-Class-3.0"):GetLookupTable()
local LibTourist = LibStub("LibTourist-3.0")
local ACR = LibStub("AceConfigRegistry-3.0")
local ACD = LibStub("AceConfigDialog-3.0")


--------------------------------------
--- Copy global functions to local ---
--------------------------------------
local pairs = pairs
local format, lower, upper, match, gsub = string.format, string.lower, string.upper, string.match, string.gsub
local tonumber = tonumber
local GetTime = GetTime

local GetDailyQuestsCompleted = GetDailyQuestsCompleted
local GetDifficultyColor = GetDifficultyColor

local UnitName = UnitName
local UnitClass = UnitClass

local GetQuestLink = GetQuestLink
local GetQuestLogLeaderBoard = GetQuestLogLeaderBoard
local GetQuestLogQuestText = GetQuestLogQuestText
local GetQuestLogSelection = GetQuestLogSelection
local GetQuestLogTitle = GetQuestLogTitle
local GetNumQuestLeaderBoards = GetNumQuestLeaderBoards
local GetNumQuestLogEntries = GetNumQuestLogEntries
local SelectQuestLogEntry = SelectQuestLogEntry

-----------------------
--- local variables ---
-----------------------
tomQuest2.playerName = UnitName("player")
local _, playerClass = UnitClass("player")
local localClass = BC[lower(playerClass):gsub("^%l", upper)]
local c = RAID_CLASS_COLORS[playerClass]
local playerColor = format("%02x%02x%02x", c.r*255, c.g*255, c.b*255)
local firstRun = true
local objectsPattern = '^' .. QUEST_OBJECTS_FOUND:gsub('(%%%d?$?.)', '(.+)') .. '$' --QUEST_OBJECTS_FOUND = "%s: %d/%d"
local monstersPattern = '^' .. QUEST_MONSTERS_KILLED:gsub('(%%%d?$?.)', '(.+)') .. '$' --QUEST_MONSTERS_KILLED = "%s slain: %d/%d"
local factionPattern = '^' .. QUEST_FACTION_NEEDED:gsub('(%%%d?$?.)', '(.+)') .. '$' --QUEST_FACTION_NEEDED = "%s: %s / %s"
local playersPattern = '^'.. QUEST_PLAYERS_KILLED:gsub('(%%%d?$?.)', '(.+)') .. '$' --QUEST_PLAYERS_KILLED
local DUNGEON = LFG_TYPE_DUNGEON
local HEROIC = DUNGEON_DIFFICULTY2
local reactions = {
	[FACTION_STANDING_LABEL1] = 1,
	[FACTION_STANDING_LABEL1_FEMALE] = 1,
	[FACTION_STANDING_LABEL2] = 2,
	[FACTION_STANDING_LABEL2_FEMALE] = 2,
	[FACTION_STANDING_LABEL3] = 3,
	[FACTION_STANDING_LABEL3_FEMALE] = 3,
	[FACTION_STANDING_LABEL4] = 4,
	[FACTION_STANDING_LABEL4_FEMALE] = 4,
	[FACTION_STANDING_LABEL5] = 5,
	[FACTION_STANDING_LABEL5_FEMALE] = 5,
	[FACTION_STANDING_LABEL6] = 6,
	[FACTION_STANDING_LABEL6_FEMALE] = 6,
	[FACTION_STANDING_LABEL7] = 7,
	[FACTION_STANDING_LABEL7_FEMALE] = 7,
	[FACTION_STANDING_LABEL8] = 8,
	[FACTION_STANDING_LABEL8_FEMALE] = 8,
}
local shortTags = {
	[ELITE] = '+',
	[GROUP] = 'g',
	[PVP] = 'p',
	[RAID] = 'r',
	[DUNGEON] = 'd',
	[HEROIC] = 'd+',
	[DAILY] = "\226\128\162",
}
local tagWeight = {
	[ELITE] = 1,
	[GROUP] = 2,
	[PVP] = 3,
	[RAID] = 4,
	[DUNGEON] = 5,
	[HEROIC] = 6,
	[DAILY] = 7,
}
tomQuest2.numCompletedQuests = 0
tomQuest2.numCompletedDailyQuests = 0
tomQuest2.numEntries, tomQuest2.numQuests = 0, 0
tomQuest2.numPartyMembers = 0
tomQuest2.party = {}
tomQuest2.quests = {}
tomQuest2.questsObjectives = {}
tomQuest2.questsItems = {}
tomQuest2.questsMobs = {}
tomQuest2.questsByZone = {}
tomQuest2.questsZones = {}
tomQuest2.questsIdToUid = {}
tomQuest2.questsTitleToUid = {}

tomQuest2.cbh = LibStub("CallbackHandler-1.0"):New(tomQuest2)
---------------
--- Options ---
---------------
local defaults = {
	profile = {
		modules = {
			syncParty = true,
			questsTracker = true,
			questsTooltip = true,
			tooltips = true,
			notifications = true,
			soundNotifications = true,
			tagQuestLog = true,
			tagQuestGivers = true,
			dailyTooltip = true,
		},
	},
}

local options = {
	name = "tomQuest2",
	handler = tomQuest2,
	type = "group",
	args = {
		modules = {
			type = "group",
			name = LC["Modules"],
			desc = LC["Modules"],
			args = {}
		}
	},--args options
} --options

----------------------
--- Initialization ---
----------------------
local prdb
tomQuest2:SetDefaultModuleState(false)
function tomQuest2:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("tomQuest2DB", defaults, "Default")
	prdb = self.db.profile
	ACR:RegisterOptionsTable("tomQuest2", options)
	ACD:AddToBlizOptions("tomQuest2", "tomQuest2")
	for k, v in self:IterateModules() do
		options.args.modules.args[k] = {
			type = "group",
			name = (v.modName or k),
			args = nil
		}
		local t
		if v.getOptions then
			t = v:getOptions()
			t.settingsHeader = {
				type = "header",
				name = LC["Settings"],
				order = 4
			}
		end
		t = t or {}
		t.toggle = {
			type = "toggle", 
			name = format(LC["Enable %s"], v.modName),
			width = "double",
			desc = v.Info, 
			order = 3,
			get = function()
				return prdb.modules[k]
			end,
			set = function(info, value)
				prdb.modules[k] = value
				if value == true then
 					tomQuest2:EnableModule(k)
				else
 					tomQuest2:DisableModule(k)
				end
			end
		}
		t.header = {
			type = "header",
			name = v.modName or k,
			order = 1
		}
		options.args.modules.args[k].args = t
	end
	self:RegisterBucketEvent("QUEST_LOG_UPDATE", 1)
	self:RegisterEvent("PLAYER_LEVEL_UP")
end

function tomQuest2:OnEnable()
	for k, v in self:IterateModules() do
		if prdb.modules[k] ~= false then
			v:Enable()
		end
	end
end
-----------------------
--- Tools Functions ---
-----------------------
function tomQuest2:getColorFromCompletionRGB(percent)
	if percent <= 0 then
		return 1, 0, 0
	elseif percent >= 1 then
		return 0, 1, 0
	elseif percent == 0.5 then
		return 1, 1, 0
	elseif percent < 0.5 then
		percent = percent / 0.5
		return 1, percent, 0
	elseif percent < 1 then
		percent = (percent-0.5) / 0.5
		return 1 - percent, 1, 0
	end
end
function tomQuest2:getColorFromCompletionHEX(percent)
	if percent <= 0 then
		return "ff0000"
	elseif percent >= 1 then
		return "00ff00"
	elseif percent == 0.5 then
		return "ffff00"
	elseif percent < 0.5 then
		percent = percent / 0.5
		return format("%02x%02x%02x",255, percent*255, 0)
	elseif percent < 1 then
		percent = (percent-0.5) / 0.5
		return format("%02x%02x%02x",255 - 255*percent, 255, 0)
	end
end
--------------------------
--- Status Computation ---
--------------------------
function tomQuest2:computeQuestStatus(uid)
	tomQuest2.quests[uid].pqStatus = ""
	local numPlayers = 0
	for partyName in pairs(tomQuest2.quests[uid].party) do
		if partyName ~= tomQuest2.playerName and (tomQuest2.quests[uid].party[partyName].complete == 1 or tomQuest2.quests[uid].party[partyName].complete == -1) then
			numPlayers = numPlayers + 1
			if numPlayers > 1 then
				tomQuest2.quests[uid].pqStatus = format("%s, |cff%s%s|r: %s",tomQuest2.quests[uid].pqStatus, tomQuest2.party[partyName].color or "ffffff", partyName:sub(1, 3), tomQuest2.quests[uid].party[partyName].complete == 1 and LC["Finished"] or LC["Failed"])
			else
				tomQuest2.quests[uid].pqStatus = format("|cff%s%s|r: %s", tomQuest2.party[partyName].color or "ffffff", partyName:sub(1, 3), tomQuest2.quests[uid].party[partyName].complete == 1 and LC["Finished"] or LC["Failed"])
			end
		end
	end
	if tomQuest2.quests[uid].party[tomQuest2.playerName] and (tomQuest2.quests[uid].party[tomQuest2.playerName].complete == 1 or tomQuest2.quests[uid].party[tomQuest2.playerName].complete == -1) then
		numPlayers = numPlayers + 1
		if numPlayers > 1 then
			tomQuest2.quests[uid].pqStatus = format("%s, %s", tomQuest2.quests[uid].pqStatus, tomQuest2.quests[uid].party[tomQuest2.playerName].complete == 1 and LC["Finished"] or LC["Failed"])
		else
			tomQuest2.quests[uid].pqStatus = format("%s", tomQuest2.quests[uid].party[tomQuest2.playerName].complete == 1 and LC["Finished"] or LC["Failed"])
		end
	end
end
function tomQuest2:computeQuestObjectiveStatus(uid, objective)
	tomQuest2.questsObjectives[uid][objective].qoStatus = ""
	local totalGot = 0
	local numPlayers = 0
	for partyName in pairs(tomQuest2.questsObjectives[uid][objective].party) do
		if partyName ~= tomQuest2.playerName then
			numPlayers = numPlayers + 1
			totalGot = totalGot + tomQuest2.questsObjectives[uid][objective].party[partyName].nGot
			if numPlayers > 1 then
				tomQuest2.questsObjectives[uid][objective].qoStatus = format("%s, |cff%s%s|r: |cff%s%s|r",tomQuest2.questsObjectives[uid][objective].qoStatus, tomQuest2.party[partyName].color, partyName:sub(1,3), tomQuest2:getColorFromCompletionHEX(tomQuest2.questsObjectives[uid][objective].party[partyName].nGot / tomQuest2.questsObjectives[uid][objective].nNeed), tomQuest2.questsObjectives[uid][objective].party[partyName].nGot == tomQuest2.questsObjectives[uid][objective].nNeed and LC["d"] or tomQuest2.questsObjectives[uid][objective].party[partyName].got)
			else
				tomQuest2.questsObjectives[uid][objective].qoStatus = format("|cff%s%s|r: |cff%s%s|r", tomQuest2.party[partyName].color, partyName:sub(1,3), tomQuest2:getColorFromCompletionHEX(tomQuest2.questsObjectives[uid][objective].party[partyName].nGot / tomQuest2.questsObjectives[uid][objective].nNeed), tomQuest2.questsObjectives[uid][objective].party[partyName].nGot == tomQuest2.questsObjectives[uid][objective].nNeed and LC["d"] or tomQuest2.questsObjectives[uid][objective].party[partyName].got)
			end
		end
	end
	if tomQuest2.questsObjectives[uid][objective].party[tomQuest2.playerName] then
		numPlayers = numPlayers + 1
		totalGot = totalGot + tomQuest2.questsObjectives[uid][objective].party[tomQuest2.playerName].nGot
		if numPlayers > 1 then
			tomQuest2.questsObjectives[uid][objective].qoStatus = format("%s, |cff%s%s|r", tomQuest2.questsObjectives[uid][objective].qoStatus, tomQuest2:getColorFromCompletionHEX(tomQuest2.questsObjectives[uid][objective].party[tomQuest2.playerName].nGot/tomQuest2.questsObjectives[uid][objective].nNeed), tomQuest2.questsObjectives[uid][objective].party[tomQuest2.playerName].nGot == tomQuest2.questsObjectives[uid][objective].nNeed and LC["d"] or tomQuest2.questsObjectives[uid][objective].party[tomQuest2.playerName].got)		
		else
			tomQuest2.questsObjectives[uid][objective].qoStatus = format("|cff%s%s|r", tomQuest2:getColorFromCompletionHEX(tomQuest2.questsObjectives[uid][objective].party[tomQuest2.playerName].nGot/tomQuest2.questsObjectives[uid][objective].nNeed), tomQuest2.questsObjectives[uid][objective].party[tomQuest2.playerName].nGot == tomQuest2.questsObjectives[uid][objective].nNeed and LC["d"] or tomQuest2.questsObjectives[uid][objective].party[tomQuest2.playerName].got)
		end
		if tomQuest2.questsObjectives[uid][objective].party[tomQuest2.playerName].nGot == tomQuest2.questsObjectives[uid][objective].nNeed then
			tomQuest2.questsObjectives[uid][objective].playerComplete = true
		else
			tomQuest2.questsObjectives[uid][objective].playerComplete = false
		end	
	end
	if totalGot == tomQuest2.quests[uid].playersOnQuest * tomQuest2.questsObjectives[uid][objective].nNeed then
		tomQuest2.questsObjectives[uid][objective].partyComplete = true
	else
		tomQuest2.questsObjectives[uid][objective].partyComplete = false
	end
	tomQuest2.questsObjectives[uid][objective].qoColor = tomQuest2:getColorFromCompletionHEX(totalGot/ (tomQuest2.quests[uid].playersOnQuest * tomQuest2.questsObjectives[uid][objective].nNeed))
	tomQuest2.questsObjectives[uid][objective].qoStatus = format("%s/|cff%s%s|r", tomQuest2.questsObjectives[uid][objective].qoStatus, tomQuest2.questsObjectives[uid][objective].qoColor, tomQuest2.questsObjectives[uid][objective].need)
end


------------------------
--- Quests Functions ---
------------------------
function tomQuest2:addZone(zone)
	tomQuest2.questsZones[zone] = {}
	if zone == LC["PARTYQUESTSZONE"] then
		tomQuest2.questsZones[zone].minLevel, tomQuest2.questsZones[zone].maxLevel, tomQuest2.questsZones[zone].level = "", "", ""
		tomQuest2.questsZones[zone].color = "ffffff"
	elseif zone == localClass then
		tomQuest2.questsZones[zone].minLevel, tomQuest2.questsZones[zone].maxLevel, tomQuest2.questsZones[zone].level = "", "", ""
		tomQuest2.questsZones[zone].color = playerColor
	else
		tomQuest2.questsZones[zone].minLevel, tomQuest2.questsZones[zone].maxLevel = LibTourist:GetLevel(zone)
		if tomQuest2.questsZones[zone].minLevel == 0 then
			tomQuest2.questsZones[zone].level = ""
			tomQuest2.questsZones[zone].color = "ffffff"
		else
			tomQuest2.questsZones[zone].level = format("(%s-%s)", tomQuest2.questsZones[zone].minLevel, tomQuest2.questsZones[zone].maxLevel)
			local r, g, b = LibTourist:GetLevelColor(zone)
			tomQuest2.questsZones[zone].color = format("%02x%02x%02x", r*255, g*255, b*255)
		end
	end
	tomQuest2.questsByZone[zone] = {}
	tomQuest2.questsByZone[zone].numQuests = 0
	tomQuest2.questsByZone[zone].quests = {}
end

function tomQuest2:addQuest(sender, first, uid, title, shortTag, level, numObjectives, complete, zone, id)
	if not tomQuest2.quests[uid] then
		tomQuest2.quests[uid] = {}
		tomQuest2.quests[uid].title = title
		tomQuest2.quests[uid].level = level
		local qc = GetDifficultyColor(level)
		tomQuest2.quests[uid].color = format("%02x%02x%02x", qc.r*255, qc.g*255, qc.b*255)
		tomQuest2.quests[uid].shortTag = shortTag
		tomQuest2.quests[uid].taggedName = format("%s%s", (level or shortTag) and format("[%s%s]", level or "", shortTag or "") or "", title)
		tomQuest2.quests[uid].zone = zone
		tomQuest2.quests[uid].playersOnQuest = 0
		tomQuest2.quests[uid].numObjectives = numObjectives
		if numObjectives > 0 then
			tomQuest2.questsObjectives[uid] = {}
		end
		tomQuest2.quests[uid].party = {}
		tomQuest2.questsByZone[zone].quests[uid] = true
		tomQuest2.questsByZone[zone].numQuests = tomQuest2.questsByZone[zone].numQuests + 1
	end
	if id ~= nil then tomQuest2.quests[uid].id = id end
	tomQuest2.quests[uid].party[sender] = {}
	tomQuest2.quests[uid].party[sender].complete = complete
	tomQuest2.quests[uid].playersOnQuest = tomQuest2.quests[uid].playersOnQuest + 1
	if sender == tomQuest2.playerName then
		if not tomQuest2.quests[uid].description or not tomQuest2.quests[uid].objective then
			SelectQuestLogEntry(id)
			tomQuest2.quests[uid].description, tomQuest2.quests[uid].objective = GetQuestLogQuestText()
		end
		if tomQuest2.quests[uid].zone == LC["PARTYQUESTSZONE"] then
			tomQuest2.quests[uid].zone = zone
			tomQuest2.questsByZone[zone].quests[uid] = true
			tomQuest2.questsByZone[zone].numQuests = tomQuest2.questsByZone[zone].numQuests + 1
			tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].quests[uid] = nil
			tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numQuests = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numQuests - 1
			if tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numQuests == 0 then
				tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]] = nil
				for k in pairs(tomQuest2.questsZones[LC["PARTYQUESTSZONE"]]) do
					tomQuest2.questsZones[LC["PARTYQUESTSZONE"]] = nil
				end
				tomQuest2.questsZones[LC["PARTYQUESTSZONE"]] = nil
			end
		end
		tomQuest2.cbh:Fire("addPlayerQuest", first, "q", uid, title, shortTag, level, numObjectives or 0, complete or 0, zone)
	end
	tomQuest2:computeQuestStatus(uid)
end

function tomQuest2:questComplete(sender, first, uid)
	tomQuest2.quests[uid].party[sender].complete = 1
	tomQuest2:computeQuestStatus(uid)
	if sender == tomQuest2.playerName then
		tomQuest2.cbh:Fire("playerQuestComplete", first, "c", uid)
	else
		tomQuest2.cbh:Fire("partyQuestComplete", sender, uid)
	end
end

function tomQuest2:questFailed(sender, first, uid)
	tomQuest2.quests[uid].party[sender].complete = -1
	tomQuest2:computeQuestStatus(uid)
	if sender == tomQuest2.playerName then
		tomQuest2.cbh:Fire("playerQuestFailed", first, "f", uid)
	else
		tomQuest2.cbh:Fire("partyQuestFailed", sender, uid)
	end
end

function tomQuest2:addQuestObjective(sender, first, uid, objective, desc, qType, got, need, nGot, nNeed)
	if not tomQuest2.questsObjectives[uid][objective] then
		tomQuest2.questsObjectives[uid][objective] = {}
		tomQuest2.questsObjectives[uid][objective].desc = desc
		tomQuest2.questsObjectives[uid][objective].qType = qType
		tomQuest2.questsObjectives[uid][objective].nNeed = nNeed
		tomQuest2.questsObjectives[uid][objective].need = need
		tomQuest2.questsObjectives[uid][objective].party = {}
		if qType == "item" or qType == "object" then
			if not tomQuest2.questsItems[desc] then
				tomQuest2.questsItems[desc] = {}
				tomQuest2.questsItems[desc].numQuests = 0
				tomQuest2.questsItems[desc].quests = {}
			end
			if not tomQuest2.questsItems[desc].quests[uid] then
				tomQuest2.questsItems[desc].numQuests = tomQuest2.questsItems[desc].numQuests + 1
				tomQuest2.questsItems[desc].quests[uid] = objective
			end
		elseif qType == "monster" then
			if not tomQuest2.questsMobs[desc] then
				tomQuest2.questsMobs[desc] = {}
				tomQuest2.questsMobs[desc].numQuests = 0
				tomQuest2.questsMobs[desc].quests = {}
			end
			if not tomQuest2.questsMobs[desc].quests[uid] then
				tomQuest2.questsMobs[desc].numQuests = tomQuest2.questsMobs[desc].numQuests + 1
				tomQuest2.questsMobs[desc].quests[uid] = objective
			end
		end
	end
	tomQuest2.questsObjectives[uid][objective].party[sender] = {}
	tomQuest2.questsObjectives[uid][objective].party[sender].nGot = nGot
	tomQuest2.questsObjectives[uid][objective].party[sender].got = got
	tomQuest2:computeQuestObjectiveStatus(uid, objective)
	if sender == tomQuest2.playerName then
		tomQuest2.cbh:Fire("addPlayerQuestObjective", first, "o", uid, objective, desc, qType, got, need, nGot, nNeed)
	elseif nGot ~= 0 then
		tomQuest2.cbh:Fire("addPartyQuestObjective", uid)
	end
end

function tomQuest2:questObjectiveProgress(sender, first, uid, objective, got, nGot)
	tomQuest2.questsObjectives[uid][objective].party[sender].nGot = nGot
	tomQuest2.questsObjectives[uid][objective].party[sender].got = got
	tomQuest2:computeQuestObjectiveStatus(uid, objective)
	if sender == tomQuest2.playerName then
		tomQuest2.cbh:Fire("playerQuestObjectiveProgress", first, "p", uid, objective, got, nGot)
	else
		tomQuest2.cbh:Fire("partyQuestObjectiveProgress", sender, uid, objective, got, nGot)
	end
end

function tomQuest2:questLost(sender, first, uid)
	tomQuest2.quests[uid].playersOnQuest = tomQuest2.quests[uid].playersOnQuest - 1
	if tomQuest2.quests[uid].numObjectives > 0 then
		for objective in pairs(tomQuest2.questsObjectives[uid]) do
			tomQuest2.questsObjectives[uid][objective].party[sender] = nil
			if tomQuest2.quests[uid].playersOnQuest == 0 then
				if (tomQuest2.questsObjectives[uid][objective].qType == "item" or tomQuest2.questsObjectives[uid][objective].qType == "object") and tomQuest2.questsItems[tomQuest2.questsObjectives[uid][objective].desc] then
					tomQuest2.questsItems[tomQuest2.questsObjectives[uid][objective].desc].numQuests = tomQuest2.questsItems[tomQuest2.questsObjectives[uid][objective].desc].numQuests - 1
					tomQuest2.questsItems[tomQuest2.questsObjectives[uid][objective].desc].quests[uid] = nil
					if tomQuest2.questsItems[tomQuest2.questsObjectives[uid][objective].desc].numQuests == 0 then
						tomQuest2.questsItems[tomQuest2.questsObjectives[uid][objective].desc].quests = nil
						tomQuest2.questsItems[tomQuest2.questsObjectives[uid][objective].desc] = nil
					end
				elseif tomQuest2.questsObjectives[uid][objective].qType == "monster" and tomQuest2.questsMobs[tomQuest2.questsObjectives[uid][objective].desc] then
					tomQuest2.questsMobs[tomQuest2.questsObjectives[uid][objective].desc].numQuests = tomQuest2.questsMobs[tomQuest2.questsObjectives[uid][objective].desc].numQuests - 1
					tomQuest2.questsMobs[tomQuest2.questsObjectives[uid][objective].desc].quests[uid] = nil
					if tomQuest2.questsMobs[tomQuest2.questsObjectives[uid][objective].desc].numQuests == 0 then
						tomQuest2.questsMobs[tomQuest2.questsObjectives[uid][objective].desc].quests = nil
						tomQuest2.questsMobs[tomQuest2.questsObjectives[uid][objective].desc] = nil
					end
				end
				tomQuest2.questsObjectives[uid][objective].party = nil
				tomQuest2.questsObjectives[uid][objective] = nil
			else
				tomQuest2:computeQuestObjectiveStatus(uid, objective)
			end
		end
	end
	tomQuest2.quests[uid].party[sender] = nil
	if tomQuest2.quests[uid].playersOnQuest == 0 then
		tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numQuests = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numQuests - 1
		tomQuest2.questsByZone[tomQuest2.quests[uid].zone].quests[uid] = nil
		if tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numQuests == 0 then
			tomQuest2.questsByZone[tomQuest2.quests[uid].zone] = nil
			tomQuest2.questsZones[tomQuest2.quests[uid].zone] = nil
		end
		if tomQuest2.quests[uid].numObjectives > 0 then
			tomQuest2.questsObjectives[uid] = nil
		end
		tomQuest2.quests[uid] = nil
	else
		if sender == tomQuest2.playerName then
			tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numQuests = tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numQuests - 1
			tomQuest2.questsByZone[tomQuest2.quests[uid].zone].quests[uid] = nil
			if tomQuest2.questsByZone[tomQuest2.quests[uid].zone].numQuests == 0 then
				tomQuest2.questsByZone[tomQuest2.quests[uid].zone] = nil
				tomQuest2.questsZones[tomQuest2.quests[uid].zone] = nil
			end
			tomQuest2.quests[uid].zone = LC["PARTYQUESTSZONE"]
			if not tomQuest2.questsZones[LC["PARTYQUESTSZONE"]] then
				tomQuest2:addZone(LC["PARTYQUESTSZONE"])
			end
			tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numQuests = tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].numQuests + 1
			tomQuest2.questsByZone[LC["PARTYQUESTSZONE"]].quests[uid] = true
		end
		tomQuest2:computeQuestStatus(uid)
	end
	if sender == tomQuest2.playerName then
		tomQuest2.cbh:Fire("playerQuestLost", first, "l", uid)
	else
		tomQuest2.cbh:Fire("partyQuestLost", uid)
	end
end


------------------------
--- QUEST_LOG_UPDATE ---
------------------------
function tomQuest2:QUEST_LOG_UPDATE()
	for i in pairs(tomQuest2.questsIdToUid) do tomQuest2.questsIdToUid[i] = nil end
	for i in pairs(tomQuest2.questsTitleToUid) do tomQuest2.questsTitleToUid[i] = nil end
	local selectionId = GetQuestLogSelection()
	local timeStamp = GetTime()
	tomQuest2.numCompletedQuests = 0
	tomQuest2.numEntries, tomQuest2.numQuests = GetNumQuestLogEntries()
	tomQuest2.numCompletedDailyQuests = GetDailyQuestsCompleted()
	if tomQuest2.numEntries > 0 then
		local zone
		for id = 1, tomQuest2.numEntries do
			local title, level, tag, group, header, collapsed, complete, daily = GetQuestLogTitle(id)
			if header then
				zone = title
				if not tomQuest2.questsZones[zone] then
					tomQuest2:addZone(zone)
				end
			else
				local uid = tonumber(match(GetQuestLink(id), 'quest:(%d+)'))
				tomQuest2.questsIdToUid[id] = uid
				tomQuest2.questsTitleToUid[title] = uid
				if not tomQuest2.quests[uid] or (tomQuest2.quests[uid] and not tomQuest2.quests[uid].party[tomQuest2.playerName]) then
					tag = tag and tag or daily and DAILY or nil
					local shortTag = format("%s%s%s", tag and shortTags[tag] or "", (group and group > 0) and group or "", daily and (tag ~= DAILY and shortTags[DAILY] or "") or "")
					local numObjectives = tomQuest2.quests[uid] and tomQuest2.quests[uid].numObjectives or GetNumQuestLeaderBoards(id)
					tomQuest2:addQuest(tomQuest2.playerName, firstRun, uid, title, shortTag, level, numObjectives or 0, complete or 0, zone, id)
				else
					if tomQuest2.quests[uid].party[tomQuest2.playerName].complete ~= complete then
						if complete == 1 then
							tomQuest2:questComplete(tomQuest2.playerName, firstRun, uid)
							tomQuest2.numCompletedQuests = tomQuest2.numCompletedQuests + 1
						elseif complete == -1 then
							tomQuest2:questFailed(tomQuest2.playerName, firstRun, uid)
						end
					elseif complete == 1 then
						tomQuest2.numCompletedQuests = tomQuest2.numCompletedQuests + 1
					end
					tomQuest2.quests[uid].id = id
				end
				if tomQuest2.quests[uid].numObjectives > 0 then
					for objective = 1, tomQuest2.quests[uid].numObjectives do
						local need, got, nNeed, nGot, desc
						local description, qType, done = GetQuestLogLeaderBoard(objective, id)
						if qType == "item" or qType == "object" then
							desc, got, need = description:match(objectsPattern)
							nNeed, nGot = tonumber(need), tonumber(got)
						elseif qType == "player" then
							desc, got, need = description:match(playersPattern)
							nNeed, nGot = tonumber(need), tonumber(got)
						elseif qType == "monster" then
							desc, got, need = description:match(monstersPattern)
							if desc == nil or got == nil or need == nil then
								desc, got, need = description:match(objectsPattern)
							end
							nNeed, nGot = tonumber(need), tonumber(got)
						elseif qType == "reputation" then
							desc, got, need = description:match(factionPattern)
							nNeed, nGot = reactions[need], reactions[got]
						elseif qType == "event" then
							 desc = description
							 got = done and 1 or 0
							 need = 1
							 nNeed, nGot = need, got
						end
						if not tomQuest2.questsObjectives[uid][objective] or (tomQuest2.questsObjectives[uid][objective] and not tomQuest2.questsObjectives[uid][objective].party[tomQuest2.playerName]) then
							tomQuest2:addQuestObjective(tomQuest2.playerName, firstRun, uid, objective, desc, qType, got, need, nGot, nNeed)
						elseif tomQuest2.questsObjectives[uid][objective].party[tomQuest2.playerName].nGot ~= nGot then
							tomQuest2.questsObjectives[uid][objective].desc = desc -- hack to make sure desc will be stored in db in case the first time GetQuestLogLeaderBoard did not return the correct desciption, maybe due to lag
							tomQuest2:questObjectiveProgress(tomQuest2.playerName, firstRun, uid, objective, got, nGot)
						else
							tomQuest2.questsObjectives[uid][objective].desc = desc -- hack to make sure desc will be stored in db in case the first time GetQuestLogLeaderBoard did not return the correct desciption, maybe due to lag
						end
					end
				end
				if not tomQuest2.quests[uid].timeStamp and not firstRun then
					tomQuest2.cbh:Fire("endPlayerQuest", "r", uid)
				end
				tomQuest2.quests[uid].timeStamp = timeStamp
			end
		end
	end
	if GetQuestLogSelection() ~= selectionId then
		SelectQuestLogEntry(selectionId)
	end
	if not firstRun then
		for uid in pairs(tomQuest2.quests) do
			if tomQuest2.quests[uid].party[tomQuest2.playerName] and tomQuest2.quests[uid].timeStamp ~= timeStamp then
				tomQuest2:questLost(tomQuest2.playerName, false, uid)
			end
		end
	else
		tomQuest2.cbh:Fire("initTrackedQuest")
		tomQuest2.cbh:Fire("initParty")
	end
	tomQuest2.cbh:Fire("updateLDBText")
	tomQuest2.cbh:Fire("updateQuestsTooltip")
	firstRun = false
end


-----------------------
--- PLAYER_LEVEL_UP ---
-----------------------
function tomQuest2:PLAYER_LEVEL_UP()
	for z in pairs(tomQuest2.questsZones) do
		if tomQuest2.questsZones[z].minLevel ~= 0 then
			local r, g, b = LibTourist:GetLevelColor(z)
			tomQuest2.questsZones[z].color = format("%02x%02x%02x", r*255, g*255, b*255)
		end
	end
	for uid in pairs(tomQuest2.quests) do
		local qc = GetDifficultyColor(tomQuest2.quests[uid].level)
		tomQuest2.quests[uid].color = format("%02x%02x%02x", qc.r*255, qc.g*255, qc.b*255)
	end
	tomQuest2.cbh:Fire("updateQuestsTooltip")
	tomQuest2.cbh:Fire("updateQuestsTracker")
end